import logging
from typing import Optional, Any

from .endpoint import Endpoint
from ..typings import SyncAsync

logger = logging.getLogger(__name__)


class NamespaceEndpoint(Endpoint):
    """ Namespace Management API """

    def create(
            self,
            custom_namespace_id: str,
            namespace_name: str,
            namespace_desc: Optional[str] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/console/namespaces",
            method="POST",
            query={
                "customNamespaceId": custom_namespace_id,
                "namespaceName": namespace_name,
                "namespaceDesc": namespace_desc
            }
        )

    def delete(
            self,
            namespace_id: str,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/console/namespaces",
            method="DELETE",
            query={
                "namespaceId": namespace_id
            }
        )

    def list(self) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/console/namespaces"
        )

    def update(
            self,
            namespace: str,
            namespace_show_name: str,
            namespace_desc: str
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/console/namespaces",
            method="PUT",
            query={
                "namespace": namespace,
                "namespaceShowName": namespace_show_name,
                "namespaceDesc": namespace_desc
            }
        )



from enum import Enum
from typing import Optional

import httpx


class HTTPResponseError(Exception):
    """Exception for HTTP errors.

    Responses from the API use HTTP response codes that are used to indicate general
    classes of success and error.
    """

    code: str = "client_response_error"
    status: int
    headers: httpx.Headers
    body: str

    def __init__(self, response: httpx.Response, message: Optional[str] = None) -> None:
        if message is None:
            message = (
                f"Request to Nacos API failed: {response.text}"
            )
        super().__init__(message)
        self.status = response.status_code
        self.headers = response.headers
        self.body = response.text



import asyncio
import json
import logging
import threading
import time
from typing import Optional, Any, Literal

from .endpoint import Endpoint
from .._chooser import Chooser
from ..typings import SyncAsync, BeatType

_ConsistencyType = Literal["ephemeral", "persist"]

logger = logging.getLogger(__name__)


class InstanceOperationMixin:

    def heartbeat(
            self,
            service_name: str,
            ip: str,
            port: int,
            weight: Optional[float] = 1.0,
            namespace_id: Optional[str] = '',
            group_name: Optional[str] = None,
            ephemeral: Optional[bool] = True,
            interval: Optional[int] = 1_000,
            skip_exception: Optional[bool] = True,
            **kwargs
    ) -> SyncAsync[Any]:
        stop_event = threading.Event()
        stop_event.cancel = stop_event.set

        def _heartbeat():
            while not stop_event.is_set():
                time.sleep(interval / 1_000)
                try:
                    self.beat(
                        service_name=service_name,
                        ip=ip,
                        port=port,
                        weight=weight,
                        namespace_id=namespace_id,
                        group_name=group_name,
                        ephemeral=ephemeral,
                        **kwargs
                    )
                except Exception as exc:
                    logger.error("Heartbeat error: %s", exc)
                    if skip_exception:
                        continue
                    raise exc

        thread = threading.Thread(target=_heartbeat)
        thread.start()
        return stop_event

    def get_one_healthy(
            self,
            service_name: str,
            namespace_id: Optional[str] = None,
            group_name: Optional[str] = None,
            clusters: Optional[str] = None,

    ):
        """ Get a healthy instance """
        instances = self.list(
            service_name=service_name,
            namespace_id=namespace_id,
            group_name=group_name,
            clusters=clusters,
            healthy_only=True
        )
        hosts = [
            (host, host.get('weight'))
            for host in instances["hosts"]
        ]
        chooser = Chooser(hosts)
        chooser.refresh()
        return chooser.random_with_weight()


class InstanceAsyncOperationMixin:

    async def heartbeat(
            self,
            service_name: str,
            ip: str,
            port: int,
            weight: Optional[float] = 1.0,
            namespace_id: Optional[str] = '',
            group_name: Optional[str] = None,
            ephemeral: Optional[bool] = True,
            interval: Optional[int] = 1_000,
            skip_exception: Optional[bool] = True,
            **kwargs
    ) -> SyncAsync[Any]:
        stop_event = threading.Event()
        stop_event.cancel = stop_event.set

        async def _async_heartbeat():
            while True:
                await asyncio.sleep(interval / 1_000)
                try:
                    await self.beat(
                        service_name=service_name,
                        ip=ip,
                        port=port,
                        weight=weight,
                        namespace_id=namespace_id,
                        group_name=group_name,
                        ephemeral=ephemeral,
                        **kwargs
                    )
                except asyncio.CancelledError:
                    break
                except Exception as exc:
                    logger.error("Heartbeat error: %s", exc)
                    if skip_exception:
                        continue
                    raise exc

        return asyncio.create_task(_async_heartbeat())

    async def get_one_healthy(
            self,
            service_name: str,
            namespace_id: Optional[str] = None,
            group_name: Optional[str] = None,
            clusters: Optional[str] = None,

    ):
        """ Get a healthy instance """
        instances = await self.list(
            service_name=service_name,
            namespace_id=namespace_id,
            group_name=group_name,
            clusters=clusters,
            healthy_only=True
        )
        hosts = [
            (host, host.get('weight'))
            for host in instances["hosts"]
        ]
        chooser = Chooser(hosts)
        chooser.refresh()
        return chooser.random_with_weight()


class _BaseInstanceEndpoint(Endpoint):
    """ Instance Management API """

    def register(
            self,
            service_name: str,
            ip: str,
            port: int,
            namespace_id: Optional[str] = '',
            weight: Optional[float] = 1.0,
            enabled: Optional[bool] = True,
            healthy: Optional[bool] = True,
            metadata: Optional[str] = None,
            cluster_name: Optional[str] = None,
            group_name: Optional[str] = None,
            ephemeral: Optional[bool] = None,
    ) -> SyncAsync[Any]:
        """ Register instance """

        return self.client.request(
            "/nacos/v1/ns/instance",
            method="POST",
            query={
                "ip": ip,
                "port": port,
                "serviceName": service_name,
                "namespaceId": namespace_id,
                "weight": weight,
                "enabled": enabled,
                "healthy": healthy,
                "metadata": metadata,
                "clusterName": cluster_name,
                "groupName": group_name,
                "ephemeral": ephemeral,
            }
        )

    def delete(
            self,
            service_name: str,
            ip: str,
            port: str,
            group_name: Optional[str] = None,
            cluster_name: Optional[str] = None,
            namespace_id: Optional[str] = None,
            ephemeral: Optional[bool] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/ns/instance",
            method="DELETE",
            query={
                "serviceName": service_name,
                "ip": ip,
                "port": port,
                "groupName": group_name,
                "clusterName": cluster_name,
                "namespaceId": namespace_id,
                "ephemeral": ephemeral,
            }
        )

    def list(
            self,
            service_name: str,
            namespace_id: Optional[str] = None,
            group_name: Optional[str] = None,
            clusters: Optional[str] = None,
            healthy_only: Optional[bool] = False,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/ns/instance/list",
            query={
                "serviceName": service_name,
                "namespaceId": namespace_id,
                "groupName": group_name,
                "clusters": clusters,
                "healthyOnly": healthy_only,
            }
        )

    def update(
            self,
            service_name: str,
            ip: str,
            port: int,
            namespace_id: Optional[str] = None,
            weight: Optional[float] = None,
            enabled: Optional[bool] = None,
            metadata: Optional[dict] = None,
            cluster_name: Optional[str] = None,
            group_name: Optional[str] = None,
            ephemeral: Optional[bool] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/ns/instance",
            method="PUT",
            query={
                "ip": ip,
                "port": port,
                "serviceName": service_name,
                "namespaceId": namespace_id,
                "weight": weight,
                "enabled": enabled,
                "metadata": metadata,
                "clusterName": cluster_name,
                "groupName": group_name,
                "ephemeral": ephemeral,
            }
        )

    def get(
            self,
            service_name: str,
            ip: str,
            port: int,
            namespace_id: Optional[str] = '',
            group_name: Optional[str] = None,
            cluster: Optional[str] = None,
            healthy_only: Optional[bool] = False,
            ephemeral: Optional[bool] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/ns/instance",
            query={
                "serviceName": service_name,
                "ip": ip,
                "port": port,
                "namespaceId": namespace_id,
                "groupName": group_name,
                "cluster": cluster,
                "healthyOnly": healthy_only,
                "ephemeral": ephemeral,
            }
        )

    def beat(
            self,
            service_name: str,
            ip: str,
            port: int,
            weight: Optional[float] = 1.0,
            namespace_id: Optional[str] = '',
            group_name: Optional[str] = None,
            ephemeral: Optional[bool] = None,
            **kwargs
    ) -> SyncAsync[Any]:
        # see: https://github.com/alibaba/nacos/issues/10448#issuecomment-1538178112
        serverName = f"{group_name}@@{service_name}" if group_name else service_name
        beat_params: BeatType = {
            "serviceName": serverName,
            "ip": ip,
            "port": port,
            "weight": weight,
            "ephemeral": ephemeral,
            **kwargs
        }
        return self.client.request(
            "/nacos/v1/ns/instance/beat",
            method="PUT",
            query={
                "serviceName": serverName,
                "beat": json.dumps(beat_params),
                "namespaceId": namespace_id,
                "groupName": group_name,
            }
        )

    def update_health(
            self,
            service_name: str,
            ip: str,
            port: int,
            healthy: bool,
            namespace_id: Optional[str] = '',
            group_name: Optional[str] = None,
            cluster_name: Optional[str] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/ns/health/instance",
            method="PUT",
            query={
                "serviceName": service_name,
                "ip": ip,
                "port": port,
                "healthy": healthy,
                "namespaceId": namespace_id,
                "groupName": group_name,
                "clusterName": cluster_name,
            }
        )

    def batch_update_metadata(
            self,
            service_name: str,
            namespace_id: str,
            metadata: dict,
            consistency_type: Optional["_ConsistencyType"] = None,
            instances: Optional[list] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/ns/instance/metadata/batch",
            method="PUT",
            query={
                "serviceName": service_name,
                "namespaceId": namespace_id,
                "metadata": json.dumps(metadata),
                "consistencyType": consistency_type,
                "instances": json.dumps(instances),
            }
        )

    def batch_delete_metadata(
            self,
            service_name: str,
            namespace_id: str,
            metadata: dict,
            consistency_type: Optional["_ConsistencyType"] = None,
            instances: Optional[list] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/ns/instance/metadata/batch",
            method="DELETE",
            query={
                "serviceName": service_name,
                "namespaceId": namespace_id,
                "metadata": json.dumps(metadata),
                "consistencyType": consistency_type,
                "instances": json.dumps(instances),
            }
        )


class InstanceEndpoint(_BaseInstanceEndpoint, InstanceOperationMixin):
    ...


class InstanceAsyncEndpoint(_BaseInstanceEndpoint, InstanceAsyncOperationMixin):
    ...



import logging
from typing import Optional, Any

from .endpoint import Endpoint
from ..typings import SyncAsync

logger = logging.getLogger(__name__)


class ServiceEndpoint(Endpoint):
    """ Service Management API """

    def create(
            self,
            service_name: str,
            namespace_id: Optional[str] = None,
            group_name: Optional[str] = None,
            protect_threshold: Optional[float] = 0,
            metadata: Optional[str] = None,
            selector: Optional[str] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/ns/service",
            method="POST",
            query={
                "serviceName": service_name,
                "namespaceId": namespace_id,
                "groupName": group_name,
                "protectThreshold": protect_threshold,
                "metadata": metadata,
                "selector": selector,
            }
        )

    def delete(
            self,
            service_name: str,
            namespace_id: Optional[str] = None,
            group_name: Optional[str] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/ns/service",
            method="DELETE",
            query={
                "serviceName": service_name,
                "groupName": group_name,
                "namespaceId": namespace_id
            }
        )

    def list(
            self,
            page_no: Optional[int] = 1,
            page_size: Optional[int] = 20,
            namespace_id: Optional[str] = None,
            group_name: Optional[str] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/ns/service/list",
            query={
                "pageNo": page_no,
                "pageSize": page_size,
                "namespaceId": namespace_id,
                "groupName": group_name,
            }
        )

    def update(
            self,
            service_name: str,
            namespace_id: Optional[str] = None,
            group_name: Optional[str] = None,
            protect_threshold: Optional[float] = None,
            metadata: Optional[str] = None,
            selector: Optional[str] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/ns/service",
            method="PUT",
            query={
                "serviceName": service_name,
                "namespaceId": namespace_id,
                "groupName": group_name,
                "protectThreshold": protect_threshold,
                "metadata": metadata,
                "selector": selector,
            }
        )

    def get(
            self,
            service_name: str,
            namespace_id: Optional[str] = '',
            group_ame: Optional[str] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/ns/service",
            query={
                "serviceName": service_name,
                "namespaceId": namespace_id,
                "groupName": group_ame,
            }
        )



# created by gpt-4

import random


class Chooser:
    def __init__(self, host_with_weight: list):
        self.host_with_weight = host_with_weight
        self.items = []
        self.weights = []

    def refresh(self):
        origin_weight_sum = 0.0
        # Preparing the valid items list and calculating the original weights sum
        for item, weight in self.host_with_weight:
            if weight <= 0:
                continue
            if float('inf') == weight:
                weight = 10000.0
            elif float('nan') == weight:
                weight = 1.0
            origin_weight_sum += weight
            self.items.append(item)

        if not self.items:
            return

        # Computing the exact weights for each item
        exact_weights = [weight / origin_weight_sum for _, weight in self.host_with_weight if weight > 0]

        # Initializing the cumulative weights array
        random_range = 0.0
        for single_weight in exact_weights:
            random_range += single_weight
            self.weights.append(random_range)

        # Checking the final weight
        double_precision_delta = 0.0001
        if abs(self.weights[-1] - 1) < double_precision_delta:
            return
        raise ValueError("Cumulative Weight calculate wrong, the sum of probabilities does not equal 1.")

    def random_with_weight(self):
        # Generating a random number between 0 and 1
        random_value = random.random()

        # Using binary search to find the index for the random value
        index = self._find_index(self.weights, random_value)

        return self.items[index]

    def _find_index(self, weights, value):
        # Perform a binary search manually since weights are not just keys
        low = 0
        high = len(weights) - 1
        while low <= high:
            mid = (low + high) // 2
            if weights[mid] < value:
                low = mid + 1
            elif weights[mid] > value:
                high = mid - 1
            else:
                return mid  # This is the exact match case
        return low  # This is the case where value should be inserted



import json
import os


class BaseCache:
    def set(self, key, value):
        raise NotImplementedError

    def get(self, key):
        raise NotImplementedError

    def exists(self, key):
        raise NotImplementedError


class MemoryCache(BaseCache):
    def __init__(self):
        self.storage = {}

    def set(self, key, value):
        self.storage[key] = value

    def get(self, key):
        return self.storage.get(key)

    def exists(self, key):
        return key in self.storage


class FileCache(BaseCache):
    def __init__(self, file_path=None):
        self.file_path = file_path or '_nacos_config_cache.json'
        if not os.path.exists(self.file_path):
            with open(self.file_path, 'w') as f:
                json.dump({}, f)

    def _read_file(self):
        with open(self.file_path, 'r') as f:
            return json.load(f)

    def _write_file(self, data):
        with open(self.file_path, 'w') as f:
            json.dump(data, f)

    def set(self, key, value):
        data = self._read_file()
        data[key] = value
        self._write_file(data)

    def get(self, key):
        data = self._read_file()
        return data.get(key)

    def exists(self, key):
        data = self._read_file()
        return key in data


memory_cache = MemoryCache()



from .config import ConfigEndpoint, ConfigAsyncEndpoint
from .instance import InstanceEndpoint, InstanceAsyncEndpoint
from .service import ServiceEndpoint
from .namespace import NamespaceEndpoint

__all__ = [
    "ConfigEndpoint",
    "ConfigAsyncEndpoint",
    "InstanceEndpoint",
    "InstanceAsyncEndpoint",
    "ServiceEndpoint",
    "NamespaceEndpoint"
]



import asyncio
import hashlib
import json
import logging
import threading
from typing import Optional, Any, Callable

import httpx

from .endpoint import Endpoint
from ..cache import BaseCache, MemoryCache, memory_cache
from ..exception import HTTPResponseError
from ..typings import SyncAsync

logger = logging.getLogger(__name__)


def _get_md5(content: Any):
    string_content = str(content) if not isinstance(content, str) else content
    return hashlib.md5(string_content.encode('utf-8')).hexdigest() if content else ''


def _get_config_key(data_id: str, group: str, tenant: str):
    # because `#` is illegal character in Nacos
    return '#'.join([data_id, group, tenant])


def _parse_config_key(key: str):
    return key.split('#')


class _BaseConfigEndpoint(Endpoint):

    def _get(
            self,
            data_id: str,
            group: str,
            tenant: Optional[str] = '',
            *,
            serialized: Optional[bool] = False
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/cs/configs",
            query={
                "dataId": data_id,
                "group": group,
                "tenant": tenant,
            },
            serialized=serialized
        )

    def publish(
            self,
            data_id: str,
            group: str,
            content: str,
            tenant: Optional[str] = '',
            type: Optional[str] = None,
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/cs/configs",
            method="POST",
            body={
                "dataId": data_id,
                "group": group,
                "tenant": tenant,
                "content": content,
                "type": type,
            }
        )

    def delete(
            self,
            data_id: str,
            group: str,
            tenant: Optional[str] = '',
    ) -> SyncAsync[Any]:
        return self.client.request(
            "/nacos/v1/cs/configs",
            method="DELETE",
            query={
                "dataId": data_id,
                "group": group,
                "tenant": tenant,
            }
        )

    @staticmethod
    def _format_listening_configs(
            data_id: str,
            group: str,
            content_md5: Optional[str] = None,
            tenant: Optional[str] = ''
    ) -> str:
        return u'\x02'.join([data_id, group, content_md5 or "", tenant]) + u'\x01'

    def subscriber(
            self,
            data_id: str,
            group: str,
            content_md5: Optional[str] = None,
            tenant: Optional[str] = '',
            timeout: Optional[int] = 30_000,
    ) -> SyncAsync[Any]:
        listening_configs = self._format_listening_configs(
            data_id, group, content_md5, tenant
        )
        return self.client.request(
            "/nacos/v1/cs/configs/listener",
            method="POST",
            body={
                "Listening-Configs": listening_configs
            },
            headers={
                "Long-Pulling-Timeout": f"{timeout}",
            },
            timeout=timeout
        )


class ConfigOperationMixin:

    @staticmethod
    def _config_callback(callback, config, serialized):
        if not callable(callback):
            return

        if serialized:
            config = json.loads(config)
        callback(config)

    def get(
            self,
            data_id: str,
            group: str,
            tenant: Optional[str] = '',
            *,
            serialized: Optional[bool] = False,
            cache: Optional[BaseCache] = None,
            default: Optional[str] = None
    ) -> SyncAsync[Any]:
        cache = cache or memory_cache
        config_key = _get_config_key(data_id, group, tenant)
        try:
            config = self._get(data_id, group, tenant, serialized=serialized)
            # todo: this function need to be optimized
            cache.set(config_key, config)
            return config
        except (httpx.ConnectError, httpx.TimeoutException) as exc:
            logger.error("Failed to get config from server, try to get from cache. %s", exc)
            return cache.get(config_key)
        except HTTPResponseError as exc:
            logger.debug("Failed to get config from server. %s", exc)
            if exc.status == 404 and default is not None:
                return default
            raise

    def subscribe(
            self,
            data_id: str,
            group: str,
            tenant: Optional[str] = '',
            timeout: Optional[int] = 30_000,
            serialized: Optional[bool] = False,
            cache: Optional[BaseCache] = None,
            callback: Optional[Callable] = None
    ) -> SyncAsync[Any]:
        cache = cache or MemoryCache()
        config_key = _get_config_key(data_id, group, tenant)
        last_md5 = _get_md5(cache.get(config_key) or '')
        stop_event = threading.Event()
        stop_event.cancel = stop_event.set

        def _subscriber():
            nonlocal last_md5
            while not stop_event.is_set():
                try:
                    response = self.subscriber(data_id, group, last_md5, tenant, timeout)
                    if not response:
                        continue
                    logging.info("Configuration update detected.")
                    last_config = self.get(data_id, group, tenant, serialized=False)
                    last_md5 = _get_md5(last_config)
                    cache.set(config_key, last_config)
                    self._config_callback(callback, last_config, serialized)
                except Exception as exc:
                    logging.error(exc)
                    stop_event.wait(1)

        thread = threading.Thread(target=_subscriber)
        thread.start()
        return stop_event


class ConfigAsyncOperationMixin:

    @staticmethod
    async def _config_callback(callback, config, serialized):
        if not callable(callback):
            return

        if serialized:
            config = json.loads(config)
        if asyncio.iscoroutinefunction(callback):
            await callback(config)
        else:
            callback(config)

    async def get(
            self,
            data_id: str,
            group: str,
            tenant: Optional[str] = '',
            *,
            serialized: Optional[bool] = False,
            cache: Optional[BaseCache] = None,
            default: Optional[str] = None
    ) -> SyncAsync[Any]:
        cache = cache or memory_cache
        config_key = _get_config_key(data_id, group, tenant)
        try:
            config = await self._get(data_id, group, tenant, serialized=serialized)
            cache.set(config_key, config)
            return config
        except (httpx.ConnectError, httpx.TimeoutException) as exc:
            logger.error("Failed to get config from server, try to get from cache. %s", exc)
            return cache.get(config_key)
        except HTTPResponseError as exc:
            logger.debug("Failed to get config from server. %s", exc)
            if exc.status == 404 and default is not None:
                return default
            raise

    async def subscribe(
            self,
            data_id: str,
            group: str,
            tenant: Optional[str] = '',
            timeout: Optional[int] = 30_000,
            serialized: Optional[bool] = False,
            cache: Optional[BaseCache] = None,
            callback: Optional[Callable] = None,
    ) -> SyncAsync[Any]:
        cache = cache or MemoryCache()
        config_key = _get_config_key(data_id, group, tenant)
        last_md5 = _get_md5(cache.get(config_key) or '')
        stop_event = threading.Event()
        stop_event.cancel = stop_event.set

        async def _async_subscriber():
            nonlocal last_md5
            while True:
                try:
                    response = await self.subscriber(
                        data_id, group, last_md5, tenant, timeout
                    )
                    if not response:
                        continue
                    logging.info("Configuration update detected.")
                    last_config = await self.get(data_id, group, tenant, serialized=False)
                    last_md5 = _get_md5(last_config)
                    cache.set(config_key, last_config)
                    await self._config_callback(callback, last_config, serialized)
                except asyncio.CancelledError:
                    break
                except Exception as exc:
                    logging.error(exc)
                    await asyncio.sleep(1)

        return asyncio.create_task(_async_subscriber())


class ConfigEndpoint(_BaseConfigEndpoint, ConfigOperationMixin):
    ...


class ConfigAsyncEndpoint(_BaseConfigEndpoint, ConfigAsyncOperationMixin):
    ...



from typing import Union

import httpx


def is_async_client(client: Union[httpx.Client, httpx.AsyncClient]):
    """ Check if the client is async client """
    return isinstance(client, httpx.AsyncClient)



import logging
from abc import abstractmethod
from json import JSONDecodeError
from typing import Union, Any, Optional, Dict, List, Generator

import httpx
from httpx import Request, Response, Auth, HTTPTransport, AsyncHTTPTransport

from .endpoints import (
    ConfigEndpoint, InstanceEndpoint, ServiceEndpoint, NamespaceEndpoint,
    ConfigAsyncEndpoint, InstanceAsyncEndpoint
)
from .exception import HTTPResponseError
from .typings import SyncAsync

logger = logging.getLogger(__name__)

DEFAULT_SERVER_ADDR = "http://localhost:8848/"
DEFAULT_NAMESPACE = ""


class BaseClient:

    def __init__(
            self,
            client: Union[httpx.Client, httpx.AsyncClient],
            server_addr: Optional[str] = None,
            username: Optional[str] = None,
            password: Optional[str] = None,
            namespace_id: Optional[str] = None,
    ):
        self.server_addr = server_addr or DEFAULT_SERVER_ADDR
        self.username = username
        self.password = password
        self.namespace_id = namespace_id or DEFAULT_NAMESPACE

        self._clients: List[Union[httpx.Client, httpx.AsyncClient]] = []
        self.client = client
        # endpoints
        self.config = ConfigEndpoint(self)
        self.instance = InstanceEndpoint(self)
        self.service = ServiceEndpoint(self)
        self.namespace = NamespaceEndpoint(self)

    @property
    def client(self) -> Union[httpx.Client, httpx.AsyncClient]:
        return self._clients[-1]

    @client.setter
    def client(self, client: Union[httpx.Client, httpx.AsyncClient]) -> None:
        client.base_url = httpx.URL(self.server_addr)
        client.timeout = httpx.Timeout(timeout=60_000 / 1_000)
        client.headers = httpx.Headers(
            {
                "User-Agent": "use-py/use-nacos",
            }
        )
        self._clients.append(client)

    def _build_request(
            self,
            method: str,
            path: str,
            query: Optional[Dict[Any, Any]] = None,
            body: Optional[Dict[Any, Any]] = None,
            headers: Optional[Dict[Any, Any]] = None,
            **kwargs
    ) -> Request:
        _headers = httpx.Headers()
        if headers:
            _headers.update(headers)
        return self.client.build_request(
            method, path, params=query, data=body, headers=_headers, **kwargs
        )

    def _parse_response(self, response: Response, serialized: bool) -> Any:

        if not serialized:
            return response.text
        try:
            body = response.json()
        except JSONDecodeError:
            body = response.text
        return body

    @abstractmethod
    def request(
            self,
            path: str,
            method: str,
            query: Optional[Dict[Any, Any]] = None,
            body: Optional[Dict[Any, Any]] = None,
            headers: Optional[Dict[Any, Any]] = None
    ) -> SyncAsync[Any]:
        raise NotImplementedError


class NacosClient(BaseClient):
    client: httpx.Client

    def __init__(
            self,
            server_addr: Optional[str] = None,
            username: Optional[str] = None,
            password: Optional[str] = None,
            namespace_id: Optional[str] = None,
            client: Optional[httpx.Client] = None,
            *,
            http_retries: Optional[int] = 3,
    ):
        """ Nacos Sync Client """
        client = client or httpx.Client(transport=HTTPTransport(retries=http_retries))
        super().__init__(
            client, server_addr, username, password, namespace_id
        )

    def request(
            self,
            path: str,
            method: str = "GET",
            query: Optional[Dict[Any, Any]] = None,
            body: Optional[Dict[Any, Any]] = None,
            headers: Optional[Dict[Any, Any]] = None,
            serialized: Optional[bool] = True,
            **kwargs
    ) -> Any:
        request = self._build_request(method, path, query, body, headers, **kwargs)
        try:
            response = self.client.send(
                request,
                auth=NacosAPIAuth(self.username, self.password)
            )
            response.raise_for_status()
            return self._parse_response(response, serialized)
        except httpx.HTTPStatusError as exc:
            raise HTTPResponseError(exc.response)


class NacosAsyncClient(BaseClient):
    client: httpx.AsyncClient

    def __init__(
            self,
            server_addr: Optional[str] = None,
            username: Optional[str] = None,
            password: Optional[str] = None,
            namespace_id: Optional[str] = None,
            client: Optional[httpx.AsyncClient] = None,
            *,
            http_retries: Optional[int] = 3,
    ):
        """ Nacos Async Client """
        client = client or httpx.AsyncClient(transport=AsyncHTTPTransport(retries=http_retries))
        super().__init__(client, server_addr, username, password, namespace_id)
        self.config = ConfigAsyncEndpoint(self)
        self.instance = InstanceAsyncEndpoint(self)

    async def request(
            self,
            path: str,
            method: str = "GET",
            query: Optional[Dict[Any, Any]] = None,
            body: Optional[Dict[Any, Any]] = None,
            headers: Optional[Dict[Any, Any]] = None,
            serialized: Optional[bool] = True,
            **kwargs
    ) -> Any:
        request = self._build_request(method, path, query, body, headers, **kwargs)
        try:
            response = await self.client.send(
                request,
                auth=NacosAPIAuth(self.username, self.password)
            )
            response.raise_for_status()
            return self._parse_response(response, serialized)
        except httpx.HTTPStatusError as exc:
            raise HTTPResponseError(exc.response)


class NacosAPIAuth(Auth):
    """ Attaches HTTP Nacos Authentication to the given Request object. """

    def __init__(
            self, username: str, password: str
    ) -> None:
        self.auth_params = {"username": username, "password": password}

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        request.url = request.url.copy_merge_params(params=self.auth_params)
        yield request



from typing import Awaitable, Union, TypeVar, TypedDict, Optional

T = TypeVar("T")
SyncAsync = Union[T, Awaitable[T]]


class BeatType(TypedDict):
    service_name: str
    ip: str
    port: int
    weight: int
    ephemeral: bool
    cluster: Optional[str]
    metadata: Optional[Union[dict, str]]
